Attribute VB_Name = "MWndProc"
Option Explicit

'constants that can be passed for the dwFlags parameter of WTSRegisterSessionNotification()
Public Const NOTIFY_FOR_THIS_SESSION As Integer = 0
Public Const NOTIFY_FOR_ALL_SESSIONS As Integer = 1

'message id to look for when processing the message (see sample code)
Public Const WM_WTSSESSION_CHANGE As Integer = &H2B1

'WParam values that can be received:
Public Const WTS_CONSOLE_CONNECT         As Integer = 1 'A session was connected to the console terminal.
Public Const WTS_CONSOLE_DISCONNECT      As Integer = 2 'A session was disconnected from the console terminal.
Public Const WTS_REMOTE_CONNECT          As Integer = 3 'A session was connected to the remote terminal.
Public Const WTS_REMOTE_DISCONNECT       As Integer = 4 'A session was disconnected from the remote terminal.
Public Const WTS_SESSION_LOGON           As Integer = 5 'A user has logged on to the session.
Public Const WTS_SESSION_LOGOFF          As Integer = 6 'A user has logged off the session.
Public Const WTS_SESSION_LOCK            As Integer = 7 'A session has been locked.
Public Const WTS_SESSION_UNLOCK          As Integer = 8 'A session has been unlocked.
Public Const WTS_SESSION_REMOTE_CONTROL  As Integer = 9 'A session has changed its remote controlled status.

'Private variables
Private Declare Function WTSRegisterSessionNotification Lib "wtsapi32.dll" (ByVal plhWnd As Long, ByVal pdwFlags As Long) As Integer
Private Declare Function WTSUnRegisterSessionNotification Lib "wtsapi32.dll" (ByVal plhWnd As Long) As Integer

Private Const FORMAT_MESSAGE_FROM_SYSTEM& = &H1000
Private Declare Function GetLastError& Lib "kernel32" ()
Private Declare Function FormatMessageW Lib "kernel32" (ByVal pdwFlags As Long, ByVal plSource As Long, ByVal pdwMessageId As Long, ByVal pdwLanguageId As Long, ByVal plBuffer As Long, ByVal plSize As Long, plArguments As Long) As Long

'Variables needed for subclassing
'WARNING: this code cannot handle multiple windows subclassing:
' - there's only one mlOldWindowProcAddress and miiSubclass module level variables
' - To fix that, use something like a collection, indexed on the hWnd
Private miiSubclass As ISubclass
Private mlOldWindowProcAddress    As Long

Private Const GWL_WNDPROC As Long = -4&
Private Declare Function CallWindowProc Lib "user32" Alias "CallWindowProcA" (ByVal plOldWndProc As Long, ByVal plhWnd As Long, ByVal plMsg As Long, ByVal pwParam As Long, ByVal plParam As Long) As Long
Private Declare Function SetWindowLong Lib "user32" Alias "SetWindowLongA" (ByVal plhWnd As Long, ByVal plPropIndex As Long, ByVal pdwNewValue As Long) As Long

Public Sub SubclassWindowForNotifications(ByRef piiSubclass As ISubclass, ByVal plhWnd As Long)
  'Subclass this window so we can handle notification messages
  Set miiSubclass = piiSubclass
  mlOldWindowProcAddress = SetWindowLong(plhWnd, GWL_WNDPROC, AddressOf WindowProc)
End Sub

Public Function WindowProc(ByVal plhWnd As Long, ByVal plMsg As Long, ByVal pwParam As Long, ByVal plParam As Long) As Long
  Dim lResult As Long
  
  If miiSubclass Is Nothing Then Exit Function 'foolprof just to avoid a crash
  
  'We're only interested in the WM_WTSSESSION_CHANGE message
  If plMsg = WM_WTSSESSION_CHANGE Then
    lResult = miiSubclass.WindowProc(plhWnd, plMsg, pwParam, plParam)
  Else
    lResult = CallWindowProc(mlOldWindowProcAddress, plhWnd, plMsg, pwParam, plParam)
  End If
  
  WindowProc = lResult
End Function

Public Sub UnsubclassWindowForNotifications(ByVal plhWnd As Long)
  'Unsubclass window
  Call SetWindowLong(plhWnd, GWL_WNDPROC, mlOldWindowProcAddress)
  Set miiSubclass = Nothing
End Sub

Public Function RegisterForSessionNotifications(ByVal plhWnd As Long, ByVal plWMNotifyScope As Long) As Boolean
  Dim iApiRet   As Integer
  iApiRet = WTSRegisterSessionNotification(plhWnd, plWMNotifyScope)
  If iApiRet = 0 Then
    Exit Function
  End If
  RegisterForSessionNotifications = True
End Function

Public Sub UnRegisterForSessionNotifications(ByVal plhWnd As Long)
  WTSUnRegisterSessionNotification plhWnd
End Sub

Public Function LastDllErrorMsg(Optional ByVal plErrCode As Long) As String
  Dim sBuffer         As String   ' Place where error description will be copied to.
  Dim lCopiedCt       As Long     ' Number of bytes copied to sBuffer
  Const BUFFER_SIZE As Long = 2048&
  
  If plErrCode = 0 Then                       ' no error code supplied
    plErrCode = Err.LastDllError              ' use the VB last known API error code
  Else
    plErrCode = Abs(plErrCode)                ' user supplied DLL error code
  End If
  If plErrCode = 0 Then Exit Function         ' bail if no error code
  ' prepare the buffer
  sBuffer = Space$(BUFFER_SIZE)
  ' translate the error code
  lCopiedCt = FormatMessageW(FORMAT_MESSAGE_FROM_SYSTEM, _
                         0&, plErrCode, _
                         0&, sBuffer, BUFFER_SIZE - 1&, 0&)
  LastDllErrorMsg = Trim$(sBuffer) 'trailing \0 not removed here
End Function

